"""Handle punctuation for pysyte"""

space = " "

bang = "!"
double_quote = '"'
hash_ = "#"
dollar = "$"
modulo = "%"
ampersand = "&"
apostrophe = single_quote = "'"
open_parens = "("
close_parens = ")"
asterisk = "*"
plus = "+"
comma = ","
minus = "-"
dot = full_stop = period = "."
forward_slash = "/"
zero = "0"
colon = ":"
semi_colon = ";"
less = "<"
equal = "="
greater = ">"
question = "?"
at = "@"
open_bracket = "["
backslash = "\\"
close_bracket = "]"
caret = "^"
underline = "_"
backtick = "`"
open_brace = "{"
pipe = "|"
close_brace = "}"

punctuation = f'!"#$%&{apostrophe})(*+,-./:;<=>?@[\\]^_`{open_brace}|{close_brace}'
